!MNH_LIC Copyright 1994-2024 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
! Modifications:
!  P. Wautelet 13/12/2018: extracted from mode_io.f90
!  P. Wautelet 14/12/2018: added IO_Filename_construct
!  P. Wautelet 05/03/2019: rename IO subroutines and modules
!  P. Wautelet 05/09/2019: IO_Mnhversion_get: Z-split files: to prevent serialization between files,
!                          nmnhversion is taken from the main file
!-----------------------------------------------------------------
module mode_io_tools

use modd_io, only: tfiledata

implicit none

private

public :: IO_Level2filenumber_get, IO_Rank_master_get, IO_Filename_construct
public :: IO_Dim_main_to_box, IO_Dim_box_bounds

contains

FUNCTION IO_Level2filenumber_get(k,nb_proc_io)
  !
  ! return the file number where to write the K level of data
  !
  IMPLICIT NONE
  INTEGER :: k,nb_proc_io
  INTEGER :: IO_Level2filenumber_get

  IO_Level2filenumber_get = MOD ((k-1) , nb_proc_io )

END FUNCTION IO_Level2filenumber_get

FUNCTION IO_Rank_master_get(IFILE,nb_proc,nb_proc_io,offset_rank)
  !
  ! return the proc number which must write the 'IFILE' file
  !
  IMPLICIT NONE
  INTEGER,           INTENT(IN) :: IFILE, nb_proc, nb_proc_io
  INTEGER, OPTIONAL, INTENT(IN) :: offset_rank

  INTEGER                       :: IO_Rank_master_get

  INTEGER                       :: ipas, irest

  ipas  =        nb_proc / nb_proc_io
  irest =  MOD ( nb_proc , nb_proc_io )

  IF  (ipas /= 0 ) THEN
     IO_Rank_master_get=ipas * IFILE + MIN(IFILE , irest )
  ELSE
     IO_Rank_master_get=MOD(IFILE , nb_proc )
  ENDIF

  !
  ! optional rank to shift for read test
  !
  IF (PRESENT(offset_rank)) THEN
     IF ( offset_rank .GT.0 ) IO_Rank_master_get=MOD(IO_Rank_master_get+offset_rank,nb_proc)
     IF ( offset_rank .LT.0 ) IO_Rank_master_get=MOD(nb_proc-IO_Rank_master_get+offset_rank,nb_proc)
  ENDIF

END FUNCTION IO_Rank_master_get


subroutine IO_Filename_construct(tpfile,hfilem)
  type(tfiledata),               intent(inout) :: tpfile
  character(len=:), allocatable, intent(out)   :: hfilem

  if (allocated(tpfile%cdirname)) then
    if(len_trim(tpfile%cdirname)>0) then
      hfilem = trim(tpfile%cdirname)//'/'//trim(tpfile%cname)
    else
      hfilem = trim(tpfile%cname)
    end if
  else
    hfilem = trim(tpfile%cname)
  end if

end subroutine IO_Filename_construct


SUBROUTINE IO_Dim_main_to_box( TPFIELD )
  ! Subroutine to transform the dimensions of the coordinates for the boxes
  USE MODD_FIELD, ONLY: NMNHDIM_NI, NMNHDIM_NJ, NMNHDIM_NI_U, NMNHDIM_NJ_U, NMNHDIM_NI_V, NMNHDIM_NJ_V,       &
                        NMNHDIM_LEVEL, NMNHDIM_LEVEL_W,                                                       &
                        NMNHDIM_BOX_NI, NMNHDIM_BOX_NJ, NMNHDIM_BOX_NI_U, NMNHDIM_BOX_NJ_U, NMNHDIM_BOX_NI_V, &
                        NMNHDIM_BOX_NJ_V, NMNHDIM_BOX_LEVEL, NMNHDIM_BOX_LEVEL_W,                             &
                        TFIELDMETADATA

  use mode_msg

  CLASS(TFIELDMETADATA), INTENT(INOUT) :: TPFIELD

  IF ( TPFIELD%NDIMS == 1 ) THEN
    SELECT CASE ( TPFIELD%CDIR )
      CASE ( 'XX' )
        SELECT CASE ( TPFIELD%NGRID )
          CASE( 1, 4 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI )                                  &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI
          CASE( 2, 5, 6, 8 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI_U )                                &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI_U
          CASE( 3, 7 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI_V )                                &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI_V
          CASE DEFAULT
            CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box', trim(tpfield%cmnhname)//': unsupported grid (XX)' )
        END SELECT

      CASE ( 'YY' )
        SELECT CASE ( TPFIELD%NGRID )
          CASE( 1, 4 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NJ )                                  &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NJ
          CASE( 2, 6 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NJ_U )                                &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NJ_U
          CASE( 3, 5, 7, 8 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NJ_V )                                &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NJ_V
          CASE DEFAULT
            CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box', trim(tpfield%cmnhname)//': unsupported grid (YY)' )
        END SELECT

      CASE ( 'ZZ' )
        SELECT CASE ( TPFIELD%NGRID )
          CASE( 1, 2, 3, 5 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_LEVEL )                               &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_LEVEL
          CASE( 4, 6, 7, 8 )
            IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_LEVEL_W )                             &
              CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                              trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
            TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_LEVEL_W
          CASE DEFAULT
            CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box', trim(tpfield%cmnhname)//': unsupported grid (ZZ)' )
      END SELECT

      CASE DEFAULT
        CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box', &
                        trim(tpfield%cmnhname) // ': unsupported CDIR (' // trim(TPFIELD%CDIR) // ')' )
    END SELECT
  END IF

  IF ( TPFIELD%NDIMS > 1 ) THEN !Case where NDIMS==1 has already been treated
    SELECT CASE ( TPFIELD%NGRID )
      CASE( 1, 4 )
        IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI )                                  &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
        TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI
      CASE( 2, 5, 6, 8 )
        IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI_U )                                &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
        TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI_U
      CASE( 3, 7 )
        IF ( TPFIELD%NDIMLIST(1)/= NMNHDIM_NI_V )                                &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 1st dimension' )
        TPFIELD%NDIMLIST(1) = NMNHDIM_BOX_NI_V
      CASE DEFAULT
        CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box', trim(tpfield%cmnhname)//': unsupported grid' )
    END SELECT
  END IF

  IF ( TPFIELD%NDIMS >= 2 ) THEN
    SELECT CASE ( TPFIELD%NGRID )
      CASE( 1, 4 )
        IF ( TPFIELD%NDIMLIST(2)/= NMNHDIM_NJ )                                  &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 2nd dimension' )
        TPFIELD%NDIMLIST(2) = NMNHDIM_BOX_NJ
      CASE( 2, 6 )
        IF ( TPFIELD%NDIMLIST(2)/= NMNHDIM_NJ_U )                                &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 2nd dimension' )
        TPFIELD%NDIMLIST(2) = NMNHDIM_BOX_NJ_U
      CASE( 3, 5, 7, 8 )
        IF ( TPFIELD%NDIMLIST(2)/= NMNHDIM_NJ_V )                                &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 2nd dimension' )
        TPFIELD%NDIMLIST(2) = NMNHDIM_BOX_NJ_V
    END SELECT
  END IF

  IF ( TPFIELD%NDIMS >= 3 ) THEN
    SELECT CASE ( TPFIELD%NGRID )
      CASE( 1, 2, 3, 5 )
        IF ( TPFIELD%NDIMLIST(3)/= NMNHDIM_LEVEL )                               &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 3rd dimension' )
        TPFIELD%NDIMLIST(3) = NMNHDIM_BOX_LEVEL
      CASE( 4, 6, 7, 8 )
        IF ( TPFIELD%NDIMLIST(3)/= NMNHDIM_LEVEL_W )                             &
          CALL Print_msg( NVERB_ERROR, 'IO', 'IO_Dim_main_to_box',               &
                          trim(tpfield%cmnhname) // ': unexpected 3rd dimension' )
        TPFIELD%NDIMLIST(3) = NMNHDIM_BOX_LEVEL_W
    END SELECT
  END IF

END SUBROUTINE IO_Dim_main_to_box


SUBROUTINE IO_Dim_box_bounds( TPBOX, TPFIELD, KIINF, KISUP, KJINF, KJSUP, KKINF, KKSUP, KKSIZE )
  ! Subroutine to return the box boundaries in global coordinates
  USE MODD_BAKOUT,     ONLY: LOUT_PHYSICAL_SIMPLIFIED
  USE MODD_DIM_n,      ONLY: NIMAX_ll, NJMAX_ll, NKMAX
  USE MODD_FIELD,      ONLY: NMNHDIM_LEVEL, NMNHDIM_LEVEL_W, NMNHDIM_BOX_LEVEL, NMNHDIM_BOX_LEVEL_W, TFIELDMETADATA
  USE MODD_OUT_n,      ONLY: CMAINDOMAINNAME, TOUTBOXMETADATA
  USE MODD_PARAMETERS, ONLY: JPHEXT, JPVEXT

  use mode_msg

  CLASS(TOUTBOXMETADATA),          INTENT(IN)  :: TPBOX
  CLASS(TFIELDMETADATA),           INTENT(IN)  :: TPFIELD
  INTEGER,               OPTIONAL, INTENT(OUT) :: KIINF
  INTEGER,               OPTIONAL, INTENT(OUT) :: KISUP
  INTEGER,               OPTIONAL, INTENT(OUT) :: KJINF
  INTEGER,               OPTIONAL, INTENT(OUT) :: KJSUP
  INTEGER,               OPTIONAL, INTENT(OUT) :: KKINF
  INTEGER,               OPTIONAL, INTENT(OUT) :: KKSUP
  INTEGER,               OPTIONAL, INTENT(IN)  :: KKSIZE ! Size of 3rd dimension (useful if this is not the vertical dimension)

  SELECT CASE ( TPFIELD%CDIR )
    CASE ( 'XY' )
      IF ( .NOT.PRESENT(KIINF) .OR. .NOT.PRESENT(KISUP) .OR. .NOT.PRESENT(KJINF) .OR. .NOT.PRESENT(KJSUP) ) &
        call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // ': missing needed dummy arguments' )

      KIINF = JPHEXT + TPBOX%NIINF
      IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 2 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
        ! There is one more PHYSICAL horizontal layer for u points (but the same number for physical + unphysical borders)
        KISUP = MIN( JPHEXT + TPBOX%NISUP + 1, NIMAX_ll + 2*JPHEXT )
      ELSE
        KISUP = JPHEXT + TPBOX%NISUP
      END IF

      KJINF = JPHEXT + TPBOX%NJINF
      IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 3 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
        ! There is one more PHYSICAL horizontal layer for v points (but the same number for physical + unphysical borders)
        KJSUP = MIN( JPHEXT + TPBOX%NJSUP + 1, NJMAX_ll + 2*JPHEXT )
      ELSE
        KJSUP = JPHEXT + TPBOX%NJSUP
      END IF

      IF ( TPFIELD%NDIMS > 2 ) THEN
        IF ( .NOT.PRESENT(KKINF) .OR. .NOT.PRESENT(KKSUP) ) &
          call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // ': missing needed dummy arguments' )

        IF ( TPFIELD%NDIMLIST(3) == NMNHDIM_LEVEL     .OR. TPFIELD%NDIMLIST(3) == NMNHDIM_LEVEL_W .OR. &
             TPFIELD%NDIMLIST(3) == NMNHDIM_BOX_LEVEL .OR. TPFIELD%NDIMLIST(3) == NMNHDIM_BOX_LEVEL_W  ) THEN
          KKINF = JPVEXT + TPBOX%NKINF
          IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 4 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
            ! There is one more PHYSICAL vertical layer for w points (but the same number for physical + unphysical borders)
            KKSUP = MIN( JPVEXT + TPBOX%NKSUP + 1, NKMAX + 2*JPVEXT )
          ELSE
            KKSUP = JPVEXT + TPBOX%NKSUP
          END IF
        ELSE
          IF ( .NOT.PRESENT(KKSIZE) ) &
            call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // &
                            ': 3rd size mandatory for non vertical dimension' )
          KKINF = 1
          KKSUP = KKSIZE
        END IF
      END IF

    CASE ( 'XX' )
      IF ( .NOT.PRESENT(KIINF) .OR. .NOT.PRESENT(KISUP) ) &
        call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // ': missing needed dummy arguments' )

      KIINF = JPHEXT + TPBOX%NIINF
      IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 2 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
        ! There is one more PHYSICAL horizontal layer for u points (but the same number for physical + unphysical borders)
        KISUP = MIN( JPHEXT + TPBOX%NISUP + 1, NIMAX_ll + 2*JPHEXT )
      ELSE
        KISUP = JPHEXT + TPBOX%NISUP
      END IF

    CASE ( 'YY' )
      IF ( .NOT.PRESENT(KJINF) .OR. .NOT.PRESENT(KJSUP) ) &
        call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // ': missing needed dummy arguments' )

      KJINF = JPHEXT + TPBOX%NJINF
      IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 3 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
        ! There is one more PHYSICAL horizontal layer for u points (but the same number for physical + unphysical borders)
        KJSUP = MIN( JPHEXT + TPBOX%NJSUP + 1, NJMAX_ll + 2*JPHEXT )
      ELSE
        KJSUP = JPHEXT + TPBOX%NJSUP
      END IF

    CASE ( 'ZZ' )
      IF ( .NOT.PRESENT(KKINF) .OR. .NOT.PRESENT(KKSUP) ) &
        call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', trim(tpfield%cmnhname) // ': missing needed dummy arguments' )

      KKINF = JPVEXT + TPBOX%NKINF
      IF ( TPBOX%CNAME == CMAINDOMAINNAME .AND. TPFIELD%NGRID == 4 .AND. .NOT.LOUT_PHYSICAL_SIMPLIFIED ) THEN
        ! There is one more PHYSICAL vertical layer for w points (but the same number for physical + unphysical borders)
        KKSUP = MIN( JPVEXT + TPBOX%NKSUP + 1, NKMAX + 2*JPVEXT )
      ELSE
        KKSUP = JPVEXT + TPBOX%NKSUP
      END IF

    CASE DEFAULT
      call Print_msg( NVERB_FATAL, 'IO', 'IO_Dim_box_bounds', &
                      trim(tpfield%cmnhname) // ': unsupported CDIR (' // trim(tpfield%cdir) // ')' )
  END SELECT

END SUBROUTINE IO_Dim_box_bounds

end module mode_io_tools



module mode_io_tools_mnhversion

use modd_io, only: tfiledata

use mode_msg

implicit none

private

public :: IO_Mnhversion_get, IO_Mnhversion_set

contains

subroutine IO_Mnhversion_get(tpfile)
!Compare MNHVERSION of file with current version and store it in file metadata
  use modd_conf,          only: nmnhversion
  use modd_field,         only: tfieldmetadata, TYPEINT

  use mode_io_field_read, only: IO_Field_read

  type(tfiledata), intent(inout) :: tpfile

  character(len=12)       :: ymnhversion_file,ymnhversion_curr
  integer :: imasdev,ibugfix
  integer :: iresp
  integer,dimension(3)    :: imnhversion
  type(tfieldmetadata)    :: tzfield

  call print_msg(NVERB_DEBUG,'IO','IO_Mnhversion_get','called for '//trim(tpfile%cname))

  if ( trim(tpfile%cmode) /= 'READ' ) &
    call print_msg(NVERB_FATAL,'IO','IO_Mnhversion_get',trim(tpfile%cname)// 'not opened in read mode')

  if ( .not. associated( tpfile%tmainfile ) ) then
    imnhversion(:) = 0
    !use tzfield because tfieldlist could be not initialised
    tzfield = tfieldmetadata(        &
      cmnhname   = 'MNHVERSION',     &
      cstdname   = '',               &
      clongname  = 'MesoNH version', &
      cunits     = '',               &
      cdir       = '--',             &
      ccomment   = '',               &
      ngrid      = 0,                &
      ntype      = TYPEINT,          &
      ndims      = 1,                &
      ltimedep   = .false.           )
    call IO_Field_read(tpfile,tzfield,imnhversion,iresp)
    if (iresp/=0) then
      tzfield%cmnhname   = 'MASDEV'
      tzfield%clongname  = 'MesoNH version (without bugfix)'
      tzfield%ndims      = 0
      call IO_Field_read(tpfile,tzfield,imasdev,iresp)
      if (iresp/=0) then
        call print_msg(NVERB_WARNING,'IO','IO_Mnhversion_get','unknown MASDEV version for '//trim(tpfile%cname))
      else
        if (imasdev<100) then
          imnhversion(1)=imasdev/10
          imnhversion(2)=mod(imasdev,10)
        else !for example for mnh 4.10
          imnhversion(1)=imasdev/100
          imnhversion(2)=mod(imasdev,100)
        end if
      end if
      !
      tzfield%cmnhname   = 'BUGFIX'
      tzfield%clongname  = 'MesoNH bugfix number'
      call IO_Field_read(tpfile,tzfield,ibugfix,iresp)
      if (iresp/=0) then
        call print_msg(NVERB_WARNING,'IO','IO_Mnhversion_get','unknown BUGFIX version for '//trim(tpfile%cname))
      else
        imnhversion(3)=ibugfix
      end if
    end if
    !
    write(ymnhversion_file,"( I0,'.',I0,'.',I0 )" ) imnhversion(1),imnhversion(2),imnhversion(3)
    write(ymnhversion_curr,"( I0,'.',I0,'.',I0 )" ) nmnhversion(1),nmnhversion(2),nmnhversion(3)
    !
    if ( imnhversion(1)==0 .and. imnhversion(2)==0 .and. imnhversion(3)==0 ) then
      call print_msg(NVERB_WARNING,'IO','IO_Mnhversion_get','file '//trim(tpfile%cname)//&
                    ' was written with an unknown version of MesoNH')
    else if (  imnhversion(1)< nmnhversion(1) .or. &
              (imnhversion(1)==nmnhversion(1) .and. imnhversion(2)< nmnhversion(2)) .or. &
              (imnhversion(1)==nmnhversion(1) .and. imnhversion(2)==nmnhversion(2) .and. imnhversion(3)<nmnhversion(3)) ) then
      call print_msg(NVERB_WARNING,'IO','IO_Mnhversion_get','file '//trim(tpfile%cname)//&
                      ' was written with an older version of MesoNH ('//trim(ymnhversion_file)//&
                      ' instead of '//trim(ymnhversion_curr)//')')
    else if (  imnhversion(1)> nmnhversion(1) .or. &
              (imnhversion(1)==nmnhversion(1) .and. imnhversion(2)> nmnhversion(2)) .or. &
              (imnhversion(1)==nmnhversion(1) .and. imnhversion(2)==nmnhversion(2) .and. imnhversion(3)>nmnhversion(3)) ) then
      call print_msg(NVERB_WARNING,'IO','IO_Mnhversion_get','file '//trim(tpfile%cname)//&
                    ' was written with a more recent version of MesoNH ('//trim(ymnhversion_file)//&
                    ' instead of '//trim(ymnhversion_curr)//')')
    else
      call print_msg(NVERB_DEBUG,'IO','IO_Mnhversion_get','file '//trim(tpfile%cname)//&
                      ' was written with the same version of MesoNH ('//trim(ymnhversion_curr)//')')
    end if
    !
    tpfile%nmnhversion(:) = imnhversion(:)
  else ! associated( tpfile%tmainfile )
    if ( .not. tpfile%tmainfile%lopened ) &
      call Print_msg( NVERB_FATAL, 'IO', 'IO_Mnhversion_get', 'tmainfile should be opened' )
    tpfile%nmnhversion(:) = tpfile%tmainfile%nmnhversion(:)
  end if
end subroutine IO_Mnhversion_get


subroutine IO_Mnhversion_set(tpfile)
  use modd_conf,  only: nmnhversion

  type(tfiledata), intent(inout) :: tpfile

  call print_msg(NVERB_DEBUG,'IO','IO_Mnhversion_set','called for '//trim(tpfile%cname))

  if ( trim(tpfile%cmode) /= 'WRITE' ) &
    call print_msg(NVERB_FATAL,'IO','IO_Mnhversion_set',trim(tpfile%cname)// 'not opened in write mode')

  tpfile%nmnhversion(:) = nmnhversion(:)
end subroutine IO_Mnhversion_set

end module mode_io_tools_mnhversion
